@@ -46,7 +46,7 @@ typedef struct File_t {
 	/* Relative position of previous cluster */
 	unsigned int PreviousRelCluNr;
 	direntry_t direntry;
-	int hint;
+	size_t hint;
 	struct dirCache_t *dcp;
 
 	unsigned int loopDetectRel;
@@ -82,7 +82,7 @@ direntry_t *getDirentry(Stream_t *Stream)
 static int recalcPreallocSize(File_t *This)
 {
 	size_t currentClusters, neededClusters;
-	int clus_size;
+	unsigned int clus_size;
 	int neededPrealloc;
 	Fs_t *Fs = This->Fs;
 	int r;
@@ -146,7 +146,7 @@ static unsigned int _countBlocks(Fs_t *This, unsigned int block)
 		block = fatDecode(This, block);
 		rel++;
 		if(_loopDetect(&oldrel, rel, &oldabs, block) < 0)
-			block = -1;
+			block = 1;
 	}
 	return blocks;
 }
@@ -175,7 +175,7 @@ void printFat(Stream_t *Stream)
 {
 	File_t *This = getUnbufferedFile(Stream);
 	unsigned long n;
-	int rel;
+	unsigned int rel;
 	unsigned long begin, end;
 	int first;
 
@@ -244,14 +244,14 @@ void printFatWithOffset(Stream_t *Stream, off_t offset) {
 static int normal_map(File_t *This, off_t where, size_t *len, int mode,
 						   mt_off_t *res)
 {
-	int offset;
+	unsigned int offset;
 	size_t end;
 	int NrClu; /* number of clusters to read */
 	unsigned int RelCluNr;
 	unsigned int CurCluNr;
 	unsigned int NewCluNr;
 	unsigned int AbsCluNr;
-	int clus_size;
+	unsigned int clus_size;
 	Fs_t *Fs = This->Fs;
 
 	*res = 0;
@@ -562,14 +562,14 @@ static unsigned int getAbsCluNr(File_t *This)
 	return 1;
 }
 
-static unsigned int func1(void *Stream)
+static size_t func1(void *Stream)
 {
 	DeclareThis(File_t);
 
 	return getAbsCluNr(This) ^ (long) This->Fs;
 }
 
-static unsigned int func2(void *Stream)
+static size_t func2(void *Stream)
 {
 	DeclareThis(File_t);
 
